! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!               Space-filling curve partitioning               !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

! use clever search algorithm in fd4_part_sfc_part1d_binsrch and fd4_part_sfc_part1d_parallel
#define CLEVER_SEARCH

!! Space-filling curve partitioning of an $fd4_domain.
!! Supports Hilbert and Morton SFC.
!!
!! Preprocessor options:
!!  - $FD4_PARTITIONING_TEST: Compile for testing the part1d routines in extern program (public instead of private)
!!
!! @see hilbert_mod
!! @see fd4_domain_mod
!! @author Matthias Lieber
module fd4_part_sfc_mod

  use fd4_globaldef_mod
  use fd4_domain_mod
  use fd4_iter_mod
  use fd4_mpi_mod
  use hilbert_mod
  use hilbert2d_mod
  use morton_mod
  use mrgrnk_mod
  use stack_mod
  use rbtree_fd4_block_mod
  
  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH
  
  integer(i_k), parameter, public :: FD4_PART_SFC_SIMPLE = 1
  integer(i_k), parameter, public :: FD4_PART_SFC_BINSRCH = 2
  integer(i_k), parameter, public :: FD4_PART_SFC_PARALLEL = 4
  
  integer(i_k), parameter, public :: FD4_PART_SFC_HILBERT = 128
  integer(i_k), parameter, public :: FD4_PART_SFC_MORTON  = 256
  
  ! public methods
  public :: fd4_part_new
  public :: fd4_part_sfc_weight

#ifdef FD4_PARTITIONING_TEST
  public :: fd4_part_sfc_part1d_binsrch
  public :: fd4_part_sfc_part1d_parallel
  public :: fd4_part_sfc_part1d_simple 
#endif
  
contains  

  include 'framework/fd4_domain.incf'
  include 'util/stack.incf'
  include 'util/math.incf'

  !! Search the largest value which not larger than $sumload in the array of prefix summed block weights
  !! $blkweights_ps starting at $pstart and return the index in $blkweights_ps.
  !!
  !! If compiled with $CLEVER_SEARCH, use a more sophisticated algorithm which starts searching at index $guess.
  !! Returns a new value for $guess for subsequent searches.
  !!
  !! Internal function, called by $fd4_part_sfc_part1d_binsrch and $fd4_part_sfc_part1d_parallel.
  !!
  function fd4_part_sfc_part1d_scan(nblocks, blkweights_ps, pstart, sumload, guess) result(pend)
    integer(i_k), intent(in)    :: nblocks                  !! number of blocks
    real(r8k),    intent(in)    :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    integer(i_k), intent(in)    :: pstart                   !! start index for search
    real(r8k),    intent(in)    :: sumload                  !! upper bound for search in $blkweights_ps values
    integer(i_k), intent(inout) :: guess                    !! guess for result value for clever algorithm
    integer(i_k) :: pend
    
    integer(i_k) :: i

    pend = nblocks+1
#ifdef CLEVER_SEARCH
    ! try a clever guess and then search with single steps forward or backward
    if( blkweights_ps(guess) > sumload ) then
      ! guessed too many blocks
      do i=guess-1,pstart,-1
        if( blkweights_ps(i) <= sumload ) then
          pend=i
          exit
        end if
      end do
    else
      ! guessed too few blocks (or exactly the right number of blocks - still need to check this)
      do i=guess+1,nblocks
        if( blkweights_ps(i) > sumload ) then
          pend=i-1
          exit
        end if
      end do
    end if
    ! guess number of blocks for next rank = number of blocks for this rank
    guess=MIN(pend+pend-pstart,nblocks)
#else
    do i=pstart+1,nblocks
      if( blkweights_ps(i) > sumload ) then
        pend = i - 1
        exit
      end if
    end do
#endif  
  end function fd4_part_sfc_part1d_scan

  !! Calculate a well-balanced partition for given block weigths.
  !! Note: weights have to be given as prefix sums!
  !!
  !! Algorithm:
  !!  - completely serial
  !!  - binary search for a value close to the optimal bottleneck value
  !!    - good values for $step are 4-16
  !!  - found bottleneck value is used to create partitioning
  !!  - it may happen that ranks end up having no local blocks, this is no bug
  !!    - increment steps to find better bottleneck value, this may (!) help
  !!
  !! Internal routine.
  subroutine fd4_part_sfc_part1d_binsrch(nblocks, nprocs, blkweights_ps, maxweight, steps, partition, maxload)
    integer(i_k), intent(in)  :: nblocks                  !! number of blocks
    integer(i_k), intent(in)  :: nprocs                   !! number of processes
    real(r8k),    intent(in)  :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    real(r8k),    intent(in)  :: maxweight                !! max weight in weight array
    integer(i_k), intent(in)  :: steps                    !! number of binary search steps to find good bottleneck value
    integer(i_k), intent(out) :: partition(0:nprocs)      !! output partition vector, contains start indices of partitions
    real(r8k),    intent(out) :: maxload                  !! estimation of load of max loaded process, real value is a bit smaller

    integer(i_k) :: crank, pstart, pend, step
    real(r8k) :: totalweight, targetload, lbload, ubload, avload, sumload, bottleneck
    integer(i_k) :: guess

    totalweight = blkweights_ps(nblocks)
    
    avload = totalweight / nprocs   ! average proc load = optimal balance
    lbload = MAX(avload, maxweight) ! lower bound for max proc load, best value that can be reached
    ubload = avload + maxweight     ! upper bound for max proc load, worst value
    !write(*,'(A,E12.4,E12.4,A,F7.4,F7.4)') 'lb/ub: ',lbload,ubload,' ~ ',avload/lbload,avload/ubload
    
    !! find good approximation of minimal max proc load ("bottleneck value")

    bottleneck = ubload
    do step=1,steps

      targetload = lbload * 0.5_r8k + ubload * 0.5_r8k
      guess = nblocks / nprocs

      sumload = targetload
      pstart = 0
      do crank=0,nprocs-2
        ! find partition for rank crank with load <= targetload
        pend = fd4_part_sfc_part1d_scan(nblocks, blkweights_ps, pstart, sumload, guess)
        if(pend>nblocks) exit
        pstart = MIN(pend, nblocks)
        sumload = targetload + blkweights_ps(pstart)
      end do ! crank
 
      if(sumload<totalweight) then
        ! miss
        lbload = targetload
      else
        ! hit
        bottleneck = targetload
        ubload = targetload
      end if
      
      !write(*,'(I6,E32.24,F7.4)') step,targetload,avload/targetload
      
    end do ! steps

    !! finally create partition with given maximum load ("bottleneck value")

    guess = nblocks / nprocs
    pstart = 0
    sumload = bottleneck
    do crank=0,nprocs-2
      ! find partition for rank crank with load <= bottleneck
      pend = fd4_part_sfc_part1d_scan(nblocks, blkweights_ps, pstart, sumload, guess)
      pstart = MIN(pend, nblocks)
      ! set partition vector
      partition(crank+1) = pstart+1
      sumload = bottleneck + blkweights_ps(pstart)
    end do ! crank

    partition(0) = 1
    partition(nprocs) = nblocks + 1

    ! this is actually not correct, real maxload can be a bit smaller
    maxload = bottleneck

  end subroutine fd4_part_sfc_part1d_binsrch


  !! Calculate a well-balanced partition for given block weigths.
  !! Note: weights have to be given as prefix sums!
  !!
  !! Algorithm:
  !!  - parallel
  !!  - every process tries a different bottleneck value interval
  !!    - binary search is applied within this interval
  !!    - $steps=1 disables the search, so that each rank checks one bottleneck value only
  !!  - best value found is determined with MPI_Allreduce
  !!  - partition for this bottleneck value is created
  !!
  !! Internal routine.
  subroutine fd4_part_sfc_part1d_parallel(nblocks, nprocs, blkweights_ps, maxweight, steps, mpi, partition, maxload, err)
    integer(i_k), intent(in)   :: nblocks                  !! number of blocks
    integer(i_k), intent(in)   :: nprocs                   !! number of processes
    real(r8k),    intent(in)   :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    real(r8k),    intent(in)   :: maxweight                !! max weight in weight array
    integer(i_k), intent(in)   :: steps                    !! number of binary search steps to find good bottleneck value
    type(fd4_mpi),intent(in)   :: mpi                      !! domain's MPI context
    integer(i_k), intent(out)  :: partition(0:nprocs)      !! output partition vector, contains start indices of partitions
    real(r8k),    intent(out)  :: maxload                  !! estimation of load of max loaded process, real value is a bit smaller
    integer(i_k), intent(out)  :: err                      !! error status: 0...ok

    integer(i_k) :: crank, pstart, pend, step
    real(r8k) :: totalweight, targetload, lbload, ubload, avload, sumload, bottleneck
    real(r8k) :: mybottleneck, mylbload, myubload
    integer(i_k) :: guess
    
    err = 0
    
    totalweight = blkweights_ps(nblocks)
    
    avload = totalweight / nprocs   ! average proc load = optimal balance
    lbload = MAX(avload, maxweight) ! lower bound for max proc load, best value that can be reached
    ubload = avload + maxweight     ! upper bound for max proc load, worst value
    

    !! every process starts with different [lb,ub] and checks if a partition exists with
    !! a bottleneck value within this interval

    mylbload = lbload +  mpi%rank    * (ubload-lbload) / mpi%nprocs
    myubload = lbload + (mpi%rank+1) * (ubload-lbload) / mpi%nprocs
    !write(*,'(I6,A,E12.4,E12.4,A,F7.4,F7.4)') mpi%rank,'  lb/ub: ',mylbload,myubload,' ~ ',avload/mylbload,avload/myubload
    
    !! find good approximation of minimal max proc load ("bottleneck value")

    mybottleneck = ubload
    do step=1,steps

      targetload = mylbload * 0.5_r8k + myubload * 0.5_r8k
      guess = nblocks / nprocs
      
      sumload = targetload
      pstart = 0
      do crank=0,nprocs-2
        ! find partition for rank crank with load <= targetload
        pend = fd4_part_sfc_part1d_scan(nblocks, blkweights_ps, pstart, sumload, guess)
        if(pend>nblocks) exit
        pstart = MIN(pend, nblocks)
        sumload = targetload + blkweights_ps(pstart)
      end do ! crank
 
      if(sumload<totalweight) then
        ! miss
        mylbload = targetload
      else
        ! hit
        mybottleneck = targetload
        myubload = targetload
      end if
      
      !write(*,'(I6,I6,E32.24,F7.4,L4)') mpi%rank,step,targetload,avload/targetload,sumload>=totalweight
      
    end do ! steps
    
    
    !! every process has now its best found bottleneck value in variable mybottleneck
    !! if a process found no working bottleneck, this variable is equal ubload
    !! so the min value of all mybottleneck variables is the best found value
    
    call MPI_Allreduce(mybottleneck, bottleneck, 1, MPI_R8K, MPI_MIN, mpi%comm, err)
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Allreduce failed')


    !if(mpi%rank==0) write(*,'(A,F7.4)') 'and the winner is:',avload/bottleneck

    !! finally create partition with given maximum load ("bottleneck value")

    guess = nblocks / nprocs
    pstart = 0
    sumload = bottleneck
    do crank=0,nprocs-2
      ! find partition for rank crank with load <= bottleneck
      pend = fd4_part_sfc_part1d_scan(nblocks, blkweights_ps, pstart, sumload, guess)
      pstart = MIN(pend, nblocks)
      ! set partition vector
      partition(crank+1) = pstart+1
      sumload = bottleneck + blkweights_ps(pstart)
    end do ! crank

    partition(0) = 1
    partition(nprocs) = nblocks + 1

    ! this is actually not correct, real maxload can be a bit smaller
    maxload = bottleneck

  end subroutine fd4_part_sfc_part1d_parallel


  !! Calculate a partition for given block weigths.
  !! Note: weights have to be given as prefix sums!
  !!
  !! Algorithm:
  !!  - completely serial
  !!  - very simple heuristic
  !!  - fast
  !!
  !! Internal routine.
  subroutine fd4_part_sfc_part1d_simple(nblocks, nprocs, blkweights_ps, partition, maxload)
    integer(i_k), intent(in)  :: nblocks                  !! number of blocks
    integer(i_k), intent(in)  :: nprocs                   !! number of processes
    real(r8k),    intent(in)  :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    integer(i_k), intent(out) :: partition(0:nprocs)      !! output partition vector, contains start indices of partitions
    real(r8k),    intent(out) :: maxload                  !! load of max loaded process

    integer(i_k) :: i, crank, pstart, pend
    real(r8k) :: totalweight

    totalweight = blkweights_ps(nblocks)
    maxload = 0.0_r8k

    partition(0) = 1
    partition(nprocs) = nblocks + 1

    pstart = 0
    do crank=0,nprocs-2
      pend = nblocks+1
      do i=pstart+1,nblocks
        if( blkweights_ps(i) > totalweight*(crank+1)/nprocs) then
          pend = i - 1
          exit
        end if
      end do
      ! for next proc
      pstart = MIN(pend, nblocks)
      ! set partition vector
      partition(crank+1) = pstart+1
      maxload = MAX(maxload, blkweights_ps(pstart) - blkweights_ps(partition(crank)-1))
    end do

    maxload = MAX(maxload, blkweights_ps(partition(nprocs)-1) - blkweights_ps(partition(nprocs-1)-1))

  end subroutine fd4_part_sfc_part1d_simple
  
  


  !! Create a new partition using block position. Considers block weights from domain%balance%blkweights.
  !!
  !! If $opt_blocks_changed is present and is set to false: $fd4_part_sfc_weight reuses the hilbert indexes
  !! and the ranking of blocks by the hilbert index from the last call!
  !!
  !! *TODO:*
  !!  - can we live without the blkweights? if blocks_changed==false:
  !!    - do sfc sorting in parallel before exchange weights
  !!    - put weights in 1d array, sorted by hilbert sfc
  !!
  subroutine fd4_part_sfc_weight(domain, all_blocks, all_blocks_len, local_blocks, local_blocks_len, method, sfctype, &
                                 balance, err, opt_blocks_changed)
    type(fd4_domain), intent(inout) :: domain                         !! the domain
    integer(i_k)    , intent(in)    :: all_blocks_len                 !! length of $all_blocks(3,:)
    integer(i_k)    , intent(in)    :: all_blocks(3,all_blocks_len)   !! array of positions of the blocks in the new partion
    integer(i_k)    , intent(inout) :: local_blocks(:,:)              !! array of new local partition (differs from rank to rank)
    integer(i_k)    , intent(out)   :: local_blocks_len               !! length of $local_blocks(3,:)
    integer(i_k)    , intent(in)    :: method                         !! 1D partitioning method to use
    integer(i_k)    , intent(in)    :: sfctype                        !! space-filling curve to use
    real(r4k)       , intent(out)   :: balance                        !! estimated balance of new partitioning
    integer(i_k)    , intent(out)   :: err                            !! error status: 0...ok
    logical,optional, intent(in)    :: opt_blocks_changed             !! tell me if the blocks have changed (created/removed)
    
    integer :: rank, nprocs
    integer(i_k) :: i, crank, pos(3), pstart, pend
    real(r8k) :: maxweight, weight, maxload
    logical :: blocks_changed
    real(r8k) :: blkweights_ps(0:all_blocks_len)   ! prefix sum of sfc sorted block weights
    integer(i_k) :: partition(0:domain%mpi%nprocs) ! partition vector (start indices)
    integer(i_k) :: steps
    integer(i_k), parameter :: minbinarysteps = 6
    integer(i_k) :: map2d(2)
    
    err = 0
    rank   = domain%mpi%rank
    nprocs = domain%mpi%nprocs
    blocks_changed = .true.


    if(present(opt_blocks_changed)) blocks_changed = opt_blocks_changed

    
    ! collect the block weights to domain%balance%blkweights
    call fd4_part_sfc_exch_weights(domain, err)
    FD4_L1_ASSERT(err/=0,'fd4_part_sfc_exch_weights failed')

    
    if(blocks_changed .or. sfctype /= domain%balance%last_sfctype ) then
    
      ! check if domain%balance%hidx and domain%balance%rnk are large enough, if not, deallocate
      if(associated(domain%balance%hidx)) then
        if(ubound(domain%balance%hidx,1)<all_blocks_len) then
          deallocate(domain%balance%hidx)
          FD4_L3_ASSERT(.not.associated(domain%balance%rnk),'rnk not associated')
          deallocate(domain%balance%rnk)
        end if
      end if

      ! check if domain%balance%hidx and domain%balance%rnk are allocated
      if(.not. associated(domain%balance%hidx)) then
        allocate(domain%balance%hidx(all_blocks_len))
        allocate(domain%balance%rnk(all_blocks_len))
        blocks_changed = .true.
      end if

      ! get index of each block on a Hilbert space-filling curve
      if(ANY(domain%blnum==1)) then
        ! 2D SFC
        if(domain%blnum(1)==1) map2d=(/2,3/)
        if(domain%blnum(2)==1) map2d=(/1,3/)
        if(domain%blnum(3)==1) map2d=(/1,2/)
        if(sfctype==FD4_PART_SFC_MORTON) then
          do i=1,all_blocks_len
            call morton2d_c2i(domain%hilbert_level,all_blocks(map2d,i),domain%balance%hidx(i))
          end do
        else
          do i=1,all_blocks_len
            call hilbert2d_c2i(domain%hilbert_level,all_blocks(map2d,i),domain%balance%hidx(i))
          end do
        end if
      else
        ! 3D SFC
        if(sfctype==FD4_PART_SFC_MORTON) then
          do i=1,all_blocks_len
            call morton_c2i(domain%hilbert_level,all_blocks(:,i),domain%balance%hidx(i))
          end do
        else
          do i=1,all_blocks_len
            call hilbert_c2i(domain%hilbert_level,all_blocks(:,i),domain%balance%hidx(i))
          end do
        end if
      end if

      ! rank the indices on the Hilbert curve, so that hidx(rnk(:)) is sorted
      call mrgrnk(domain%balance%hidx(1:all_blocks_len), domain%balance%rnk(1:all_blocks_len))
      
      domain%balance%last_sfctype = sfctype

    end if
    
    ! apply prefix sum to weight array
    maxweight = 0.0
    blkweights_ps(0) = 0.0
    do i=1,all_blocks_len
      pos(1:3) = all_blocks(1:3,domain%balance%rnk(i))
      weight = domain%balance%blkweights( pos(1), pos(2), pos(3) )
      blkweights_ps(i) = weight + blkweights_ps(i-1)
      maxweight = MAX(maxweight, weight)
    end do

    
    !! call partitioning routine

    if(method==FD4_PART_SFC_BINSRCH) then
      ! good serial method
      steps = minbinarysteps
      call fd4_part_sfc_part1d_binsrch(all_blocks_len, nprocs, blkweights_ps, maxweight, steps, partition, maxload)
      !
    elseif(method==FD4_PART_SFC_PARALLEL) then    
      ! parallel method
      if(nprocs < minbinarysteps**2) then
        ! raise number of steps each rank has to compute
        steps = log2i( (minbinarysteps**2+nprocs-1) / nprocs)
      else
        steps = 1
      end if
      call fd4_part_sfc_part1d_parallel(all_blocks_len, nprocs, blkweights_ps, maxweight, steps, domain%mpi, partition, maxload,err)
      !
    else
      ! simple method is default
      call fd4_part_sfc_part1d_simple(all_blocks_len, nprocs, blkweights_ps, partition, maxload)
      !
    end if

    balance = REAL( blkweights_ps(all_blocks_len) / (nprocs * maxload), r4k)

    !! finally apply partition to FD4

    domain%blkmap_last = domain%blkmap_now
    domain%blkmap_now  = 3 - domain%blkmap_now
    if(all_blocks_len<PRODUCT(domain%blnum)) domain%blkmap(:,:,:,domain%blkmap_now) = -1
    local_blocks_len = 0
    ! set block map
    do crank=0,nprocs-1
      do i=partition(crank),partition(crank+1)-1
        pos(1:3) = all_blocks(1:3,domain%balance%rnk(i))
        domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_now ) = crank
      end do
    end do
    ! set local list of blocks
    pstart = partition(rank)
    pend   = partition(rank+1)-1
    if(pstart<=pend) then
      domain%balance%localweight = REAL(blkweights_ps(pend) - blkweights_ps(pstart-1),r4k)
      local_blocks_len = pend - pstart + 1
      do i=1,local_blocks_len
        local_blocks(1:3,i) = all_blocks(1:3,domain%balance%rnk(i+pstart-1))
      end do
    else
      domain%balance%localweight = 0.0
      local_blocks_len = 0
    end if
    
    ! apply periodic boundary on blkmap (which is allocated with a boundary to make life easier later on)
    pos = domain%blnum
    if(domain%periodic(1)) then
      domain%blkmap(0,1:pos(2),1:pos(3),domain%blkmap_now) = domain%blkmap(pos(1),1:pos(2),1:pos(3),domain%blkmap_now)
      domain%blkmap(pos(1)+1,1:pos(2),1:pos(3),domain%blkmap_now) = domain%blkmap(1,1:pos(2),1:pos(3),domain%blkmap_now)
    end if
    if(domain%periodic(2)) then
      domain%blkmap(1:pos(1),0,1:pos(3),domain%blkmap_now) = domain%blkmap(1:pos(1),pos(2),1:pos(3),domain%blkmap_now)
      domain%blkmap(1:pos(1),pos(2)+1,1:pos(3),domain%blkmap_now) = domain%blkmap(1:pos(1),1,1:pos(3),domain%blkmap_now)
    end if
    if(domain%periodic(3)) then
      domain%blkmap(1:pos(1),1:pos(2),0,domain%blkmap_now) = domain%blkmap(1:pos(1),1:pos(2),pos(3),domain%blkmap_now)
      domain%blkmap(1:pos(1),1:pos(2),pos(3)+1,domain%blkmap_now) = domain%blkmap(1:pos(1),1:pos(2),1,domain%blkmap_now)
    end if      
    
    
#if ( FD4_VERBOSE_LEVEL >= 4 )
    write(fd4_out,'(A,A,I6,E13.4)') FD4_LINEHD,'new partition: ',local_blocks_len, domain%balance%localweight
#endif
   
  
  end subroutine fd4_part_sfc_weight



  !! Exchange the weight of my blocks with all other processes.
  !!
  !! Updates domain%balance%blkweights.
  !!
  !! *TODO*:
  !!  - exchange only modified block weights (with a given tolerance) to minimize communication?
  !!  - do we need the blkweights array? mpi%recvreal has all the weights.
  !! 
  !! Internal routine.
  subroutine fd4_part_sfc_exch_weights(domain, err)
    type(fd4_domain), intent(inout), target :: domain       !! the domain
    integer(i_k),     intent(out)           :: err          !! error status: 0...ok
    
    type(fd4_mpi), pointer :: mpi
    integer(i_k) :: crank, nblocks, k, j, i
    type(fd4_iter) :: iter

    mpi => domain%mpi
    
    mpi%recvcnt(:) = 0
    nblocks = 0
    err = -1
    
    ! loop over blkmap to build recv counts for MPI_Allgatherv
    ! (number of blocks for each process)
    do k=1,domain%blnum(3)
      do j=1,domain%blnum(2)
        do i=1,domain%blnum(1)
          crank = domain%blkmap(i,j,k,domain%blkmap_now)
          if( crank >= 0 ) then
            mpi%recvcnt(crank) = mpi%recvcnt(crank) + 1
          end if
        end do
      end do
    end do

    FD4_L3_ASSERT(mpi%recvcnt(mpi%rank)/=domain%blocks%length,'nblocks not as expected')
    
    ! build the recv displacements
    mpi%recvdis(0) = 0
    do crank=1,mpi%nprocs-1
      mpi%recvdis(crank) = mpi%recvcnt(crank-1) + mpi%recvdis(crank-1)
    end do

    nblocks = mpi%recvdis(mpi%nprocs-1) + mpi%recvcnt(mpi%nprocs-1)

    ! this may fail, because domain%blockcount may be inconsistent
    ! after calling fd4_util_put_array
    !FD4_L3_ASSERT(nblocks/=domain%blockcount,'nblocks not as expected')
    
    ! allocate arrays for sending and receiving the block weights
    if( size(mpi%recvreal,1) < nblocks ) then
      deallocate(mpi%recvreal)
      allocate(mpi%recvreal(nblocks))
    end if
    if( size(mpi%sendreal,1) < mpi%recvcnt(mpi%rank) ) then
      deallocate(mpi%sendreal)
      allocate(mpi%sendreal(mpi%recvcnt(mpi%rank)))
    end if
    
    ! get the local block weights and write them to mpi%sendreal
    i = 0
    iter%cur => rbtree_fd4_block_iterator_first(iter%biter, domain%blocks)
    do while(associated(iter%cur))
      i = i + 1
      mpi%sendreal(i) = iter%cur%weight
      iter%cur => rbtree_fd4_block_iterator_next(iter%biter)
    end do
    
    FD4_L3_ASSERT(i/=domain%blocks%length,'nblocks not as expected')
        

    ! exchange the block weights
    call MPI_Allgatherv (mpi%sendreal, mpi%recvcnt(mpi%rank),              MPI_R4K, &
                         mpi%recvreal, mpi%recvcnt,           mpi%recvdis, MPI_R4K, mpi%comm, err )
    err = err - MPI_SUCCESS
    FD4_L1_ASSERT(err/=0,'MPI_Allgatherv failed')
    
   
    
    ! loop over blkmap and fill domain%balance%blkweights
    ! this method assumes the loop order in this loop has the same ordering
    ! as the index used to sort the blocks in domain%blocks
    do k=1,domain%blnum(3)
      do j=1,domain%blnum(2)
        do i=1,domain%blnum(1)
          crank = domain%blkmap(i,j,k,domain%blkmap_now)
          if( crank >= 0 ) then
            domain%balance%blkweights(i,j,k) = mpi%recvreal( mpi%recvdis(crank)+1 )
            mpi%recvdis(crank) = mpi%recvdis(crank) + 1
          else
            ! set default weight
            !TODO: do this only if the block has been removed last load balancing
            !      (or block will be added this time)
            domain%balance%blkweights(i,j,k) = domain%balance%defaultweight  &
             * REAL(   ( domain%d(1)%bl(i+1) - domain%d(1)%bl(i) )           &
                     * ( domain%d(2)%bl(j+1) - domain%d(2)%bl(j) )           &
                     * ( domain%d(3)%bl(k+1) - domain%d(3)%bl(k) ), r4k )
          end if
        end do
      end do
    end do

#if ( FD4_VERBOSE_LEVEL >= 3 )
    do crank=1,mpi%nprocs-1
      if(mpi%recvdis(crank) - mpi%recvdis(crank-1) /= mpi%recvcnt(crank)) then
        FD4_L1_ERROR('inconsistency after loop over blkmap')
      end if
    end do
#endif
    
    err = 0
    
  end subroutine fd4_part_sfc_exch_weights
  



  !! Create a new partition using block position. Weights are not considered.
  !!
  !! The special option $add_only effects:
  !! - all existing blocks (in blkmap) are left where they are
  !! - new blocks are added as if the option was not set
  !! - this (intentionally) results in a non balanced partitioning, but minimizes transfer costs
  !!   when - in a second partitioning step - blocks are migrated
  !!
  !! *TODO*:
  !! - think about what happens if a rank has no neighbor (this may happen also in parallel execution!)
  subroutine fd4_part_new(domain, all_blocks, all_blocks_len, local_blocks, local_blocks_len, add_only, opt_blocks_changed)
    type(fd4_domain), intent(inout) :: domain                         !! the domain
    integer(i_k)    , intent(in)    :: all_blocks_len                 !! length of $all_blocks(3,:)
    integer(i_k)    , intent(in)    :: all_blocks(3,all_blocks_len)   !! array of positions of the blocks in the new partion
    integer(i_k)    , intent(inout) :: local_blocks(:,:)              !! array of new local partition (differs from rank to rank)
    integer(i_k)    , intent(out)   :: local_blocks_len               !! length of $local_blocks(3,:)
    logical,optional, intent(in)    :: add_only                       !! only add *new* blocks, but don't touch existing blocks
    logical,optional, intent(in)    :: opt_blocks_changed             !! tell me if the blocks have changed (created/removed)
    
    integer :: rank, nprocs
    integer(i_k) :: i, crank, psum, pos(3)
    logical :: ao
    logical :: blocks_changed
    integer(i_k) :: map2d(2)
    
    if(present(add_only)) then
      ao = add_only
    else
      ao = .false.
    end if
    
    rank   = domain%mpi%rank
    nprocs = domain%mpi%nprocs
    blocks_changed = .true.
    
    if(present(opt_blocks_changed)) blocks_changed = opt_blocks_changed
    
    ! check if domain%balance%hidx and domain%balance%rnk are large enough, if not, deallocate
    if(associated(domain%balance%hidx)) then
      if(ubound(domain%balance%hidx,1)<all_blocks_len) then
        deallocate(domain%balance%hidx)
        FD4_L3_ASSERT(.not.associated(domain%balance%rnk),'rnk not associated')
        deallocate(domain%balance%rnk)
      end if
    end if

    ! check if domain%balance%hidx and domain%balance%rnk are allocated
    if(.not. associated(domain%balance%hidx)) then
      allocate(domain%balance%hidx(all_blocks_len))
      allocate(domain%balance%rnk(all_blocks_len))
      blocks_changed = .true.
    end if

    if(blocks_changed) then
    
      ! get index of each block on a Hilbert space-filling curve
      if(ANY(domain%blnum==1)) then
        ! 2D Hilbert curve
        if(domain%blnum(1)==1) map2d=(/2,3/)
        if(domain%blnum(2)==1) map2d=(/1,3/)
        if(domain%blnum(3)==1) map2d=(/1,2/)
        do i=1,all_blocks_len
          call hilbert2d_c2i(domain%hilbert_level,all_blocks(map2d,i),domain%balance%hidx(i))
        end do
      else
        ! 3D Hilbert curve
        do i=1,all_blocks_len
          call hilbert_c2i(domain%hilbert_level,all_blocks(:,i),domain%balance%hidx(i))
        end do
      end if

      ! rank the indices on the hilbert curve, so that hidx(rnk(:)) is sorted
      call mrgrnk(domain%balance%hidx(1:all_blocks_len), domain%balance%rnk(1:all_blocks_len))

    end if
        
    
    local_blocks_len = 0
    crank = 0
    psum = 0
    domain%blkmap_last = domain%blkmap_now
    domain%blkmap_now  = 3 - domain%blkmap_now
    if(all_blocks_len<PRODUCT(domain%blnum)) domain%blkmap(:,:,:,domain%blkmap_now) = -1
    do i=1,all_blocks_len
      psum = psum + nprocs
      if(all_blocks_len*(crank+1)<psum) crank=crank+1
      pos(1:3) = all_blocks(1:3,domain%balance%rnk(i))
      if(ao .and. domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_last ) /= -1 ) then
        if(domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_last ) == domain%mpi%rank ) then
          local_blocks_len = local_blocks_len + 1
          local_blocks(:,local_blocks_len) = pos(:)
        end if
        domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_now ) = domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_last )
      else
        domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_now ) = crank
        if(rank == crank) then
          local_blocks_len = local_blocks_len + 1
          local_blocks(:,local_blocks_len) = pos(:)
        end if
      end if
    end do
    
    ! apply periodic boundary on blkmap (which is allocated with a boundary to make life easier later on)
    pos = domain%blnum
    if(domain%periodic(1)) then
      domain%blkmap(0,1:pos(2),1:pos(3),domain%blkmap_now) = domain%blkmap(pos(1),1:pos(2),1:pos(3),domain%blkmap_now)
      domain%blkmap(pos(1)+1,1:pos(2),1:pos(3),domain%blkmap_now) = domain%blkmap(1,1:pos(2),1:pos(3),domain%blkmap_now)
    end if
    if(domain%periodic(2)) then
      domain%blkmap(1:pos(1),0,1:pos(3),domain%blkmap_now) = domain%blkmap(1:pos(1),pos(2),1:pos(3),domain%blkmap_now)
      domain%blkmap(1:pos(1),pos(2)+1,1:pos(3),domain%blkmap_now) = domain%blkmap(1:pos(1),1,1:pos(3),domain%blkmap_now)
    end if
    if(domain%periodic(3)) then
      domain%blkmap(1:pos(1),1:pos(2),0,domain%blkmap_now) = domain%blkmap(1:pos(1),1:pos(2),pos(3),domain%blkmap_now)
      domain%blkmap(1:pos(1),1:pos(2),pos(3)+1,domain%blkmap_now) = domain%blkmap(1:pos(1),1:pos(2),1,domain%blkmap_now)
    end if      
    
    
#if ( FD4_VERBOSE_LEVEL >= 4 )
    write(fd4_out,'(A,A,I6)') FD4_LINEHD,'new partition: ',local_blocks_len
#endif
  
  end subroutine fd4_part_new



end module fd4_part_sfc_mod
