! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                       2D Hilbert Curve                       !
!                                                              !
!--------------------------------------------------------------!


!! Simple 2D Hilbert curve subroutines.
!! Based on a table description of the curve.
!!
!! Routines accept position and hilbert-index starting at 1.
!!
!! @author Matthias Lieber
module hilbert2d_mod

  use kinds_mod
  implicit none
  private  
  
  public :: hilbert2d_i2c, hilbert2d_c2i
  
  !! Type for one entry in the 2D Hilbert curve description table.
  !!
  !! Values of the $mv item are:
  !! - 1/-1 means x+1/x-1 
  !! - 2/-2 means y+1/y-1 
  type hilbert_table
    integer(i_k) :: mv(0:2)      !! spatial moves of the curve
    integer(i_k) :: next(0:3)    !! next states when refinining (quadrant is sequence number)
    integer(i_k) :: num2ref(0:3) !! sequence number on curve primitve to reference (spatial) quadrant 
    integer(i_k) :: ref2num(0:3) !! reference (spatial) quadrant to sequence number on curve primitve
  end type hilbert_table  
  
  
  type(hilbert_table), parameter :: tab(0:3) =                            &
  (/                                                                      &
    hilbert_table( (/ 2, 1,-2/), (/1,0,0,2/), (/0,2,3,1/), (/0,3,1,2/) ), &
    hilbert_table( (/ 1, 2,-1/), (/0,1,1,3/), (/0,1,3,2/), (/0,1,3,2/) ), &
    hilbert_table( (/-1,-2, 1/), (/3,2,2,0/), (/3,2,0,1/), (/2,3,1,0/) ), &
    hilbert_table( (/-2,-1, 2/), (/2,3,3,1/), (/3,1,0,2/), (/2,1,3,0/) )  &
   /)
  
contains


!! Get index on curve at given 2D coordinate (coordinates-to-index).
!!
!! Has logarithmic complexity: number of loop iterations = $lev.
!! Due to performance reasons, no checking of arguments is performend!
subroutine hilbert2d_c2i(lev, p, idx)

  implicit none

  integer(i_k), intent(in)  :: lev    !! level of the hilbert curve (1=2x2, 2=4x4, 3=8x8, etc)
  integer(i_k), intent(in)  :: p(2)   !! 2D coordinates [1 ... 2^$lev]
  integer(i_k), intent(out) :: idx    !! index on the 2D hilbert curve [1 ... (2^$lev)^2]
  
  integer(i_k) :: l,len,dim,q,s,rq,cp(2)
    
  cp(:)=p(:)-1
  idx=1
  s=0
  dim=2**lev
  ! length of curve in base quadrant
  len=dim**2

#ifdef MO_HILBERT_ERRCHECK  
  ! very expensive check
  if(cp(1)>=dim .or. cp(1)<0 .or. cp(2)>=dim .or. cp(2)<0) then
    print *,'hilbert2d_c2i: position out of bounds', cp
    stop
  end if
#endif

  do l=lev,1,-1

    ! dimension and length of refined quadrants
    dim=dim/2
    len=len/4
    
    ! determine spatial quadrant and reduce spatial coordinates
    ! so that they are relative to refined quadrant
    ! reference quadrant:
    ! 2 3
    ! 0 1
    rq=0
    if(cp(1)>=dim) then
      rq=ior(1,rq)
      cp(1)=cp(1)-dim
    end if
    if(cp(2)>=dim) then
      rq=ior(2,rq)
      cp(2)=cp(2)-dim
    end if
    
    !print *,'lev',l,'  s',s,'  rq',rq
    
    ! get sequence number of quadrant q in the curve primitive
    q=tab(s)%ref2num(rq)
    
    ! add offset to index    
    idx=idx+q*len
    
    ! get next state
    s=tab(s)%next(q)
    
  end do
  
end subroutine hilbert2d_c2i


!! Get 2D coordinate at given index on curve (index-to-coordinates).
!!
!! Has logarithmic complexity: number of loop iterations = $lev.
!! Due to performance reasons, no checking of arguments is performend!
subroutine hilbert2d_i2c(lev, idx, p)

  implicit none

  integer(i_k), intent(in)  :: lev   !! level of the hilbert curve (1=2x2, 2=4x4, 3=8x8, etc)
  integer(i_k), intent(in)  :: idx   !! index on the 2D hilbert curve [1 ... (2^$lev)^2 ]
  integer(i_k), intent(out) :: p(2)  !! 2D coordinates [1 ... 2^$lev]

  integer(i_k) :: l,len,dim,q,cidx,s,rq
  
  p(:)=1
  s=0
  cidx=idx-1
  dim=2**lev
  
  ! length of curve in base quadrant
  len=dim**2

#ifdef MO_HILBERT_ERRCHECK  
  if(cidx>=len .or. cidx<0) then
    print *,'hilbert2d_i2c: cidx >= len or cidx < 0: ', cidx
    stop
  end if
#endif

  do l=lev,1,-1

    ! dimension of refined quadrants
    dim=dim/2
    
    ! determine number of quadrant q in the curve primitive
    if(cidx<len/2) then
      if(cidx<len/4) then
        q=0
      else
        q=1
      end if
    else
      if(cidx<3*len/4) then
        q=2
      else
        q=3
      end if
    end if
    
    ! get actual spatial quadrant, depending on curve primitive
    ! reference quadrant:
    ! 2 3
    ! 0 1
    rq=tab(s)%num2ref(q)
    
    ! update coordinates
    if(iand(1,rq)==1) p(1)=p(1)+dim
    if(iand(2,rq)==2) p(2)=p(2)+dim
    
    ! length of curve in refined quadrant
    len=len/4
    
    ! decrease idx by offset in current quadrant
    cidx=cidx-q*len
    
    ! next state
    s=tab(s)%next(q)
    
  end do
 
end subroutine hilbert2d_i2c

end module hilbert2d_mod
